use crate::dghv::{encrypt_bit, generate_secret_key};
use crate::utils::generate_random_integer;
use rand::rngs::StdRng;
use rand::Rng;
use rand::SeedableRng;
use rug::Integer;

pub struct PublicKey {
    pub xs: Vec<Integer>,
}

pub struct PrivateKey {
    pub p: Integer,
}

pub fn generate_keys(gamma: u32, eta: u32, rho: u32, theta: usize) -> (PrivateKey, PublicKey) {
    let p: Integer = generate_secret_key(eta);

    let mut xs: Vec<Integer> = Vec::with_capacity(theta);
    for _ in 0..theta {
        let xi = encrypt_bit(0u8, &p, gamma, rho);
        xs.push(xi);
    }
    xs.sort_by(|a, b| b.cmp(a));

    while xs[0].is_even() {
        xs.remove(0);
        let xi = encrypt_bit(0u8, &p, gamma, rho);
        xs.push(xi);
        xs.sort_by(|a, b| b.cmp(a));
    }

    (PrivateKey { p }, PublicKey { xs })
}

/// c = (m + 2*r + sum(b_i * x_i) for i>0 ) mod x0
/// x0 = largest public element, the x_i are randomly chosen
pub fn encrypt_bit_asym(m: u8, pk: &PublicKey, rho: u32) -> Integer {
    assert!(m == 0 || m == 1, "Message bit must be 0 or 1");
    let mut rng = StdRng::from_os_rng();
    let x0 = &pk.xs[0];

    let num_keys = &pk.xs.len();
    let proba: f64 = 2f64 / (*num_keys as f64);

    // random x_i
    let mut sum = Integer::from(0);
    for x in pk.xs.iter().skip(1) {
        if rng.random_bool(proba) {
            sum += x;
        }
    }

    // noise
    let r = generate_random_integer(rho);
    let mut c = Integer::from(m);
    c += Integer::from(2) * r;
    c += sum;

    c %= x0;
    c
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::decrypt_bit;

    #[test]
    fn test_key_generation_properties() {
        let eta: u32 = 10000;
        let theta: usize = 5;
        let rho: u32 = 100;
        let (sk, pk) = generate_keys(rho, eta, rho, theta);

        assert_eq!(sk.p.significant_bits(), eta);
        assert!(sk.p.is_odd());

        assert_eq!(pk.xs.len(), theta);
        assert!(pk.xs[0].is_odd());

        for xi in pk.xs.iter().skip(1) {
            assert!(pk.xs[0] >= *xi);
        }
    }

    #[test]
    fn test_encrypt_decrypt_bit() {
        use rayon::prelude::*;

        let eta: u32 = 10000;
        let theta: usize = 50;
        let rho: u32 = 100;

        (0..=127).into_par_iter().for_each(|_| {
            let (sk, pk) = generate_keys(rho, eta, rho, theta);
            for &m in &[0u8, 1u8] {
                let c = encrypt_bit_asym(m, &pk, rho);
                let m2 = decrypt_bit(&c, &sk.p);
                assert_eq!(m, m2);
            }
        });
    }
}
