use crate::dghv_asym::{PrivateKey, PublicKey};
use rug::Integer;
use std::fs::{read_to_string, File};
use std::io::{self, Read, Write};
use std::path::Path;

pub fn write_privkey<P: AsRef<Path>>(prefix: P, sk: &PrivateKey) -> io::Result<()> {
    let mut file = File::create(prefix.as_ref().with_extension("privkey"))?;
    let mut hex = sk.p.to_string_radix(16);
    if hex.len() % 2 != 0 {
        hex.insert(0, '0');
    }
    writeln!(file, "{}", hex)?;
    Ok(())
}

pub fn write_pubkey<P: AsRef<Path>>(prefix: P, pk: &PublicKey) -> io::Result<()> {
    let mut file = File::create(prefix.as_ref().with_extension("pubkey"))?;
    for x in &pk.xs {
        let mut hex = x.to_string_radix(16);
        if hex.len() % 2 != 0 {
            hex.insert(0, '0');
        }
        writeln!(file, "{}", hex)?;
    }
    Ok(())
}

pub fn read_privkey<P: AsRef<Path>>(path: P) -> io::Result<PrivateKey> {
    let hex = std::fs::read_to_string(path)?;
    let hex = hex.trim();
    let p = Integer::from_str_radix(hex, 16)
        .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))?;
    Ok(PrivateKey { p })
}

pub fn read_pubkey<P: AsRef<Path>>(path: P) -> io::Result<PublicKey> {
    let mut xs = Vec::new();
    let content = {
        let mut s = String::new();
        File::open(path)?.read_to_string(&mut s)?;
        s
    };
    for line in content.lines() {
        // each line is a hex-encoded key
        let bytes =
            hex::decode(line.trim()).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))?;
        let hex = hex::encode(bytes);
        let x = Integer::from_str_radix(&hex, 16)
            .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))?;
        xs.push(x);
    }
    Ok(PublicKey { xs })
}

pub fn read_hex_from_file<P: AsRef<Path>>(path: P) -> io::Result<String> {
    let content = read_to_string(path)?;
    Ok(content.trim().to_string())
}

pub fn write_hex_to_file<P: AsRef<Path>>(path: P, hex_str: &str) -> io::Result<()> {
    let mut file = File::create(path)?;
    writeln!(file, "{}", hex_str)?;
    Ok(())
}
