mod bootstrapping;
mod dghv;
mod dghv_asym;
mod fileutils;
mod utils;
use crate::dghv::{decrypt_bit, encrypt_bit, generate_secret_key};
use crate::dghv_asym::{encrypt_bit_asym, generate_keys};
use crate::fileutils::{
    read_hex_from_file, read_privkey, read_pubkey, write_hex_to_file, write_privkey, write_pubkey,
};
use clap::{ArgGroup, Args, Parser, Subcommand};
use rug::Integer;

#[derive(Parser)]
#[command(author, version, about, long_about = None)]
#[command(propagate_version = true)]
struct Cli {
    #[command(subcommand)]
    command: Commands,
}

#[derive(Subcommand)]
enum Commands {
    /// Generate a secret key
    Keygen {
        /// security parameter eta (p bit-length)
        #[clap(short = 's', long, default_value_t = 6400)]
        size: u32,
    },
    /// Encrypt a single bit
    Encrypt(EncryptArgs),
    /// Decrypt a single bit
    Decrypt(DecryptArgs),
    /// Generate pair (pubkey, privkey)
    KeygenPair {
        /// output path prefix for keys
        #[arg(short = 'o', long)]
        path: String,
        /// security parameter eta (p bit-length)
        #[clap(short = 's', long, default_value_t = 6400)]
        size: u32,
        /// noise parameter
        #[clap(short = 'n', long, default_value_t = 80)]
        noise: u32,
        /// number of public key elements (theta)
        #[arg(short = 't', long, default_value_t = 10)]
        theta: usize,
    },
}

#[derive(Args)]
#[command(group(ArgGroup::new("keysource").required(true).args(&["key", "privkey", "pubkey"])))]
struct EncryptArgs {
    /// bit to encrypt
    #[arg(short = 'b', long)]
    bit: u8,
    /// Key as hex string
    #[arg(short = 'k', long)]
    key: Option<String>,
    /// path to private key file
    #[arg(long)]
    privkey: Option<String>,
    /// path to public key file
    #[arg(long)]
    pubkey: Option<String>,
    /// noise parameter
    #[arg(short = 'n', long, default_value_t = 80)]
    noise: u32,
    /// optional output file for ciphertext (hex-encoded)
    #[arg(short = 'o', long)]
    output: Option<String>,
}

#[derive(Args)]
#[command(group(ArgGroup::new("keysource").required(true).args(&["key", "privkey"])))]
#[command(group(ArgGroup::new("cipher_source").required(true).args(&["ciphertext", "file"])))]
struct DecryptArgs {
    /// ciphertext hex string
    #[arg(short = 'c', long)]
    ciphertext: Option<String>,
    /// path to ciphertext file
    #[arg(short = 'f', long)]
    file: Option<String>,
    /// key as hex string
    #[arg(short = 'k', long)]
    key: Option<String>,
    /// path to private key file
    #[arg(long)]
    privkey: Option<String>,
}

fn main() {
    let cli = Cli::parse();

    match cli.command {
        Commands::Keygen { size } => {
            let sk = generate_secret_key(size);
            println!("Secret key (p): {}", sk.to_string_radix(16));
        }
        Commands::Encrypt(args) => {
            let ct = if let Some(pub_path) = args.pubkey {
                let pk = read_pubkey(pub_path).expect("Failed to load public key");
                encrypt_bit_asym(args.bit, &pk, args.noise)
            } else {
                let p = if let Some(hex_str) = args.key {
                    Integer::from_str_radix(&hex_str, 16).expect("Invalid key hex")
                } else {
                    read_privkey(args.privkey.unwrap())
                        .expect("Failed to load private key")
                        .p
                };
                encrypt_bit(args.bit, &p, args.noise, args.noise)
            };
            let hexct = ct.to_string_radix(16);
            if let Some(outfile) = args.output {
                write_hex_to_file(outfile, &hexct).expect("Failed to write ciphertext to file");
            } else {
                println!("Encrypted bit (hex): {}", hexct);
            }
        }
        Commands::Decrypt(args) => {
            let ct_hex = if let Some(hex_str) = args.ciphertext {
                hex_str
            } else {
                read_hex_from_file(args.file.clone().unwrap())
                    .expect("Failed to read ciphertext file")
            };
            let ct = Integer::from_str_radix(&ct_hex, 16).expect("Invalid ciphertext");
            let p = if let Some(hex_str) = args.key {
                Integer::from_str_radix(&hex_str, 16).expect("Invalid key hex")
            } else {
                read_privkey(args.privkey.unwrap())
                    .expect("Failed to load private key")
                    .p
            };
            let res = decrypt_bit(&ct, &p);
            println!("Decrypted bit: {}", res);
        }
        Commands::KeygenPair {
            path,
            size,
            noise,
            theta,
        } => {
            let (sk, pk) = generate_keys(noise, size, noise, theta);
            write_privkey(&path, &sk).expect("Failed to write privkey");
            write_pubkey(&path, &pk).expect("Failed to write pubkey");
            println!("keys written to {}.[privkey|pubkey]", path);
        }
    }
}
